/*
 *  i2c_pca_platform.c
 *
 *  Platform driver for the PCA9564 I2C controller.
 *
 *  Copyright (C) 2008 Pengutronix
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License version 2 as
 *  published by the Free Software Foundation.

 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <linux/errno.h>
#include <linux/i2c.h>
#include <linux/interrupt.h>
#include <linux/platform_device.h>
#ifdef TAHITI
#include "i2c-algo-pca.h"
#else
#include <linux/i2c-algo-pca.h>
#endif

#include <linux/platform_data/i2c-pca-platform.h>
#include <linux/gpio.h>
#include <linux/io.h>

#include <asm/irq.h>

struct i2c_pca_pf_data {
	void __iomem			*reg_base;
	int				irq;	/* if 0, use polling */
	int				gpio;
	wait_queue_head_t		wait;
	struct i2c_adapter		adap;
	struct i2c_algo_pca_data	algo_data;
	unsigned long			io_base;
	unsigned long			io_size;
};

/* Read/Write functions for different register alignments */

static int i2c_pca_pf_readbyte8(void *pd, int reg)
{
	struct i2c_pca_pf_data *i2c = pd;
	return ioread8(i2c->reg_base + reg);
}

static int i2c_pca_pf_readbyte16(void *pd, int reg)
{
	struct i2c_pca_pf_data *i2c = pd;
	return ioread8(i2c->reg_base + reg * 2);
}

static int i2c_pca_pf_readbyte32(void *pd, int reg)
{
	struct i2c_pca_pf_data *i2c = pd;
	return ioread8(i2c->reg_base + reg * 4);
}

static void i2c_pca_pf_writebyte8(void *pd, int reg, int val)
{
	struct i2c_pca_pf_data *i2c = pd;
	iowrite8(val, i2c->reg_base + reg);
}

static void i2c_pca_pf_writebyte16(void *pd, int reg, int val)
{
	struct i2c_pca_pf_data *i2c = pd;
	iowrite8(val, i2c->reg_base + reg * 2);
}

static void i2c_pca_pf_writebyte32(void *pd, int reg, int val)
{
	struct i2c_pca_pf_data *i2c = pd;
	iowrite8(val, i2c->reg_base + reg * 4);
}


static int i2c_pca_pf_waitforcompletion(void *pd)
{
	struct i2c_pca_pf_data *i2c = pd;
	unsigned long timeout;
	long ret;

	if (i2c->irq) {
		ret = wait_event_timeout(i2c->wait,
			i2c->algo_data.read_byte(i2c, I2C_PCA_CON)
			& I2C_PCA_CON_SI, i2c->adap.timeout);
	} else {
		/* Do polling */
		timeout = jiffies + i2c->adap.timeout;
		do {
			ret = time_before(jiffies, timeout);
			if (i2c->algo_data.read_byte(i2c, I2C_PCA_CON)
					& I2C_PCA_CON_SI)
				break;
			udelay(100);
		} while (ret);
	}

	return ret > 0;
}

static void i2c_pca_pf_dummyreset(void *pd)
{
	struct i2c_pca_pf_data *i2c = pd;
	printk(KERN_WARNING "%s: No reset-pin found. Chip may get stuck!\n",
		i2c->adap.name);
}

static void i2c_pca_pf_resetchip(void *pd)
{
	struct i2c_pca_pf_data *i2c = pd;

	gpio_set_value(i2c->gpio, 0);
	ndelay(100);
	gpio_set_value(i2c->gpio, 1);
}

static irqreturn_t i2c_pca_pf_handler(int this_irq, void *dev_id)
{
	struct i2c_pca_pf_data *i2c = dev_id;
#ifdef TAHITI
    int int_status = 0, INT_OFFSET = 0x40;
    unsigned char int_mask = 0x3e; /* bit 1 - bit 5 is for i2c0 - i2c4 interrupt */
    
    /* IRQ 4,5,6,7,8 might be shared with other devices.
     * Check interrupt reigster 0x40 of FPGA, return it if it's not our interrupt. --Dirk
     */
    int_status =i2c->algo_data.read_byte(i2c, INT_OFFSET);
    if(!((unsigned char)int_status & int_mask))
		return IRQ_NONE;
#endif
    
	if ((i2c->algo_data.read_byte(i2c, I2C_PCA_CON) & I2C_PCA_CON_SI) == 0)
		return IRQ_NONE;

	wake_up(&i2c->wait);

	return IRQ_HANDLED;
}


static int i2c_pca_pf_probe(struct platform_device *pdev)
{
	struct i2c_pca_pf_data *i2c;
	struct resource *res;
	struct i2c_pca9564_pf_platform_data *platform_data =
				dev_get_platdata(&pdev->dev);
	int ret = 0;
	int irq;

	res = platform_get_resource(pdev, IORESOURCE_MEM, 0);
#ifdef TAHITI
	//irq = platform_get_irq(pdev, 0);
	irq = 0;
#else
	irq = platform_get_irq(pdev, 0);
#endif
	/* If irq is 0, we do polling. */

	if (res == NULL) {
		ret = -ENODEV;
		goto e_print;
	}
    
#ifdef TAHITI
    /* Do not lock this memory region, we have to access it from user space --Dirk */
#else
	if (!request_mem_region(res->start, resource_size(res), res->name)) {
		ret = -ENOMEM;
		goto e_print;
	}
#endif

	i2c = kzalloc(sizeof(struct i2c_pca_pf_data), GFP_KERNEL);
	if (!i2c) {
		ret = -ENOMEM;
		goto e_alloc;
	}

	init_waitqueue_head(&i2c->wait);

	i2c->reg_base = ioremap(res->start, resource_size(res));
	if (!i2c->reg_base) {
		ret = -ENOMEM;
		goto e_remap;
	}
#ifdef TAHITI
        printk("BUS %d  i2c->reg_base = %#lx, res->start = %#lx\n",
              pdev->id, (ulong)i2c->reg_base, (ulong)res->start);
#endif
	i2c->io_base = res->start;
	i2c->io_size = resource_size(res);
	i2c->irq = irq;

	i2c->adap.nr = pdev->id;
	i2c->adap.owner = THIS_MODULE;

#ifdef TAHITI
    /* Show bus number --Dirk */
	snprintf(i2c->adap.name, sizeof(i2c->adap.name),
		 "PCA9564/PCA9665 BUS %d at 0x%08lx",
		         pdev->id,
                 (unsigned long) res->start);
#else
	snprintf(i2c->adap.name, sizeof(i2c->adap.name),
		 "PCA9564/PCA9665 at 0x%08lx",
		 (unsigned long) res->start);
#endif
	i2c->adap.algo_data = &i2c->algo_data;
	i2c->adap.dev.parent = &pdev->dev;

#ifdef TAHITI
    /* Have to check clock speed -- Dirk */
#endif
	if (platform_data) {
		i2c->adap.timeout = platform_data->timeout;
		i2c->algo_data.i2c_clock = platform_data->i2c_clock_speed;
	}

	i2c->algo_data.data = i2c;
	i2c->algo_data.wait_for_completion = i2c_pca_pf_waitforcompletion;
    
#ifdef TAHITI
    /* Do not implement it yet, shall we?  --Dirk */
#endif
	i2c->algo_data.reset_chip = i2c_pca_pf_dummyreset;

#ifdef TAHITI
    /* Assume go to default --Dirk */
#endif
	switch (res->flags & IORESOURCE_MEM_TYPE_MASK) {
	case IORESOURCE_MEM_32BIT:
		i2c->algo_data.write_byte = i2c_pca_pf_writebyte32;
		i2c->algo_data.read_byte = i2c_pca_pf_readbyte32;
		break;
	case IORESOURCE_MEM_16BIT:
		i2c->algo_data.write_byte = i2c_pca_pf_writebyte16;
		i2c->algo_data.read_byte = i2c_pca_pf_readbyte16;
		break;
	case IORESOURCE_MEM_8BIT:
	default:
		i2c->algo_data.write_byte = i2c_pca_pf_writebyte8;
		i2c->algo_data.read_byte = i2c_pca_pf_readbyte8;
		break;
	}

	if (irq) {
		ret = request_irq(irq, i2c_pca_pf_handler,
#ifdef TAHITI
    /* Add IRQF_SHARED in case it's not enabled --Dirk */
			IRQF_TRIGGER_FALLING | IRQF_SHARED, pdev->name, i2c);
#else
			IRQF_TRIGGER_FALLING, pdev->name, i2c);
#endif
		if (ret)
			goto e_reqirq;
	}

#ifdef TAHITI
    /* Init PCA9564 --Dirk */
#endif
	if (i2c_pca_add_numbered_bus(&i2c->adap) < 0) {
		ret = -ENODEV;
		goto e_adapt;
	}

	platform_set_drvdata(pdev, i2c);

#ifdef TAHITI
	printk(KERN_INFO "%s BUS %d registered.\n", i2c->adap.name, pdev->id);
#else
	printk(KERN_INFO "%s registered.\n", i2c->adap.name);
#endif

	return 0;

e_adapt:
	if (irq)
		free_irq(irq, i2c);
e_reqirq:
	if (i2c->gpio > -1)
		gpio_free(i2c->gpio);

	iounmap(i2c->reg_base);
e_remap:
	kfree(i2c);
e_alloc:
	release_mem_region(res->start, resource_size(res));
e_print:
#ifdef TAHITI
	printk(KERN_ERR "Registering PCA9564/PCA9665 BUS %d FAILED! (%d)\n", pdev->id, ret);
#else
	printk(KERN_ERR "Registering PCA9564/PCA9665 FAILED! (%d)\n", ret);
#endif
	return ret;
}

static int i2c_pca_pf_remove(struct platform_device *pdev)
{
	struct i2c_pca_pf_data *i2c = platform_get_drvdata(pdev);

	i2c_del_adapter(&i2c->adap);

	if (i2c->irq)
		free_irq(i2c->irq, i2c);

	if (i2c->gpio > -1)
		gpio_free(i2c->gpio);

	iounmap(i2c->reg_base);
#ifdef TAHITI
	//release_mem_region(i2c->io_base, i2c->io_size);
#else
	release_mem_region(i2c->io_base, i2c->io_size);
#endif
	kfree(i2c);

	return 0;
}

static struct platform_driver i2c_pca_pf_driver = {
	.probe = i2c_pca_pf_probe,
	.remove = i2c_pca_pf_remove,
	.driver = {
		.name = "i2c-pca-platform",
	},
};

#ifdef TAHITI
/*
*  Netapp customize area.  
*  It's for BlackPearl IOM only, follow FPGA specification.  
*
*  --Dirk
*/
static void NTAP_I2C_BUS_RELEASE(struct device *pdev) {}

#define LPC_FPGA_DECODE_ADDR    0xfe500000       // Need to get LGMR reg. 0x98h 31:16 from CPU and replace it --Dirk
#define FPGA_I2C_0_OFF          0x200
#define FPGA_I2C_1_OFF          0x210
#define FPGA_I2C_2_OFF          0x220
#define FPGA_I2C_3_OFF          0x230
#define FPGA_I2C_4_OFF          0x240
#define FPGA_I2C_SIZE           0xf

static struct resource NTAP_I2C_BUS0_RES[] = {
    [0] = {
        .start  = LPC_FPGA_DECODE_ADDR + FPGA_I2C_0_OFF,
        .end    = LPC_FPGA_DECODE_ADDR + FPGA_I2C_0_OFF + FPGA_I2C_SIZE,
        .flags  = IORESOURCE_MEM,
    },
    [1] = {
        .start  = 5,
        .end    = 5,
        .flags  = IORESOURCE_IRQ,
    },
};

static struct resource NTAP_I2C_BUS1_RES[] = {
    [0] = {
        .start  = LPC_FPGA_DECODE_ADDR + FPGA_I2C_1_OFF,
        .end    = LPC_FPGA_DECODE_ADDR + FPGA_I2C_1_OFF + FPGA_I2C_SIZE,
        .flags  = IORESOURCE_MEM,
    },
    [1] = {
        .start  = 6,
        .end    = 6,
        .flags  = IORESOURCE_IRQ,
    },
};

static struct resource NTAP_I2C_BUS2_RES[] = {
    [0] = {
        .start  = LPC_FPGA_DECODE_ADDR + FPGA_I2C_2_OFF,
        .end    = LPC_FPGA_DECODE_ADDR + FPGA_I2C_2_OFF + FPGA_I2C_SIZE,
        .flags  = IORESOURCE_MEM,
    },
    [1] = {
        .start  = 7,
        .end    = 7,
        .flags  = IORESOURCE_IRQ,
    },
};

static struct resource NTAP_I2C_BUS3_RES[] = {
    [0] = {
        .start  = LPC_FPGA_DECODE_ADDR + FPGA_I2C_3_OFF,
        .end    = LPC_FPGA_DECODE_ADDR + FPGA_I2C_3_OFF + FPGA_I2C_SIZE,
        .flags  = IORESOURCE_MEM,
    },
    [1] = {
        .start  = 10,
        .end    = 10,
        .flags  = IORESOURCE_IRQ,
    },
};

static struct resource NTAP_I2C_BUS4_RES[] = {
    [0] = {
        .start  = LPC_FPGA_DECODE_ADDR + FPGA_I2C_4_OFF,
        .end    = LPC_FPGA_DECODE_ADDR + FPGA_I2C_4_OFF + FPGA_I2C_SIZE,
        .flags  = IORESOURCE_MEM,
    },
    [1] = {
        .start  = 11,
        .end    = 11,
        .flags  = IORESOURCE_IRQ,
    },
};

/* Bus0 - Bus2 are from PCH */
static struct platform_device NTAP_I2C_BUS0 = {
    .name           = "i2c-pca-platform",
    .id             = 3,    /* -1 is dynamaic */
    .num_resources  = ARRAY_SIZE(NTAP_I2C_BUS0_RES),
    .resource       = NTAP_I2C_BUS0_RES,
    .dev.release    = NTAP_I2C_BUS_RELEASE,
};

static struct platform_device NTAP_I2C_BUS1 = {
    .name           = "i2c-pca-platform",
    .id             = 4,
    .num_resources  = ARRAY_SIZE(NTAP_I2C_BUS1_RES),
    .resource       = NTAP_I2C_BUS1_RES,
    .dev.release    = NTAP_I2C_BUS_RELEASE,
};

static struct platform_device NTAP_I2C_BUS2 = {
    .name           = "i2c-pca-platform",
    .id             = 5,
    .num_resources  = ARRAY_SIZE(NTAP_I2C_BUS2_RES),
    .resource       = NTAP_I2C_BUS2_RES,
    .dev.release    = NTAP_I2C_BUS_RELEASE,
};

static struct platform_device NTAP_I2C_BUS3 = {
    .name           = "i2c-pca-platform",
    .id             = 6,
    .num_resources  = ARRAY_SIZE(NTAP_I2C_BUS3_RES),
    .resource       = NTAP_I2C_BUS3_RES,
    .dev.release    = NTAP_I2C_BUS_RELEASE,
};

static struct platform_device NTAP_I2C_BUS4 = {
    .name           = "i2c-pca-platform",
    .id             = 7,
    .num_resources  = ARRAY_SIZE(NTAP_I2C_BUS4_RES),
    .resource       = NTAP_I2C_BUS4_RES,
    .dev.release    = NTAP_I2C_BUS_RELEASE,
};

static struct platform_device *NTAP_I2C_MASTER[] __initdata = {
    &NTAP_I2C_BUS0,
    &NTAP_I2C_BUS1,
    &NTAP_I2C_BUS2,
    &NTAP_I2C_BUS3,
    &NTAP_I2C_BUS4,
};

static int __init NetappI2CMasterInit(void)
{
    /* Add platform device --Dirk */
    platform_add_devices(NTAP_I2C_MASTER, ARRAY_SIZE(NTAP_I2C_MASTER));

    /* Register driver for platform device */
    return platform_driver_register(&i2c_pca_pf_driver);
}

static void __exit NetappI2CMasterExit(void)
{
    platform_driver_unregister(&i2c_pca_pf_driver);

    /* Del platform device --Dirk */
    platform_device_unregister(&NTAP_I2C_BUS0);
    platform_device_unregister(&NTAP_I2C_BUS1);
    platform_device_unregister(&NTAP_I2C_BUS2);
    platform_device_unregister(&NTAP_I2C_BUS3);
    platform_device_unregister(&NTAP_I2C_BUS4);
}

module_init(NetappI2CMasterInit);
module_exit(NetappI2CMasterExit);

#else
module_platform_driver(i2c_pca_pf_driver);
#endif

MODULE_AUTHOR("Wolfram Sang <kernel@pengutronix.de>");
MODULE_DESCRIPTION("I2C-PCA9564/PCA9665 platform driver");
MODULE_LICENSE("GPL");


